"use strict";
const electron = require("electron");
const dialogs = require("simple-dialogs");
const async = require("async");
const i18n = require("../shared/i18n");
const settings = require("./settings");
const splashScreen = require("./splashScreen");
const updateManager = require("./updateManager");
const sidebar = require("./sidebar");
const me = require("./sidebar/me");
const home = require("./home");
const serverSettings = require("./serverSettings");
const serverSettingsSystems = require("./serverSettings/systems");
const tabs = require("./tabs");
const openServerSettings_1 = require("./tabs/openServerSettings");
const localServer = require("./localServer");
const chat = require("./chat");
const WelcomeDialog_1 = require("./WelcomeDialog");
electron.ipcRenderer.on("init", onInitialize);
electron.ipcRenderer.on("quit", onQuit);
const namespaces = [
    "common", "startup",
    "sidebar", "server",
    "welcome", "home"
];
function onInitialize(sender, corePath, userDataPath, languageCode) {
    settings.corePath = corePath;
    settings.userDataPath = userDataPath;
    i18n.languageCode = languageCode;
    i18n.load(namespaces, () => { settings.load(onSettingsLoaded); });
}
function onQuit() {
    serverSettings.applyScheduledSave();
    settings.applyScheduledSave();
    localServer.shutdown(() => { electron.ipcRenderer.send("ready-to-quit"); });
}
function onSettingsLoaded(err) {
    if (err != null) {
        const label = i18n.t("startup:errors.couldNotLoadSettings", { reason: err.message });
        const options = {
            validationLabel: i18n.t("startup:startAnyway"),
            cancelLabel: i18n.t("common:actions.close")
        };
        new dialogs.ConfirmDialog(label, options, (shouldProceed) => {
            if (!shouldProceed) {
                electron.remote.app.quit();
                return;
            }
            updateManager.checkForUpdates(start);
        });
        return;
    }
    updateManager.checkForUpdates(start);
}
function start() {
    sidebar.start();
    home.start();
    serverSettings.start();
    splashScreen.fadeOut(() => {
        if (settings.nickname == null) {
            async.series([showWelcomeDialog, installFirstSystem]);
        }
        else {
            me.start();
            chat.start();
            updateSystemsAndPlugins();
        }
    });
}
function showWelcomeDialog(callback) {
    new WelcomeDialog_1.default((result) => {
        settings.nickname = result.nickname;
        settings.presence = result.connectToChat ? "online" : "offline";
        settings.savedChatrooms = ["#superpowers-html5"];
        if (i18n.languageCode !== "en" && chat.languageChatRooms.indexOf(i18n.languageCode) !== -1) {
            settings.savedChatrooms.push(`#superpowers-html5-${i18n.languageCode}`);
        }
        settings.scheduleSave();
        me.start();
        chat.start();
        setTimeout(callback, 500);
    });
}
function installFirstSystem(callback) {
    const label = i18n.t("welcome:askGameInstall.prompt");
    const options = {
        header: i18n.t("welcome:askGameInstall.title"),
        validationLabel: i18n.t("common:actions.install"),
        cancelLabel: i18n.t("common:actions.skip")
    };
    new dialogs.ConfirmDialog(label, options, (installGame) => {
        if (!installGame) {
            localServer.start();
            callback();
            return;
        }
        const waitingGameInstallElt = document.querySelector(".waiting-game-install");
        async.series([
                (cb) => {
                openServerSettings_1.default();
                serverSettingsSystems.action("install", { systemId: "game" }, () => { cb(); });
                waitingGameInstallElt.hidden = false;
            },
                (cb) => {
                waitingGameInstallElt.hidden = true;
                const label = i18n.t("welcome:serverInformation.info");
                const options = {
                    haeder: i18n.t("welcome:serverInformation.title"),
                    closeLabel: i18n.t("welcome:serverInformation.gotIt")
                };
                new dialogs.InfoDialog(label, options, cb);
            },
                (cb) => {
                localServer.start();
                const label = i18n.t("welcome:sidebarInformation.info");
                const options = {
                    header: i18n.t("welcome:sidebarInformation.title")
                };
                new dialogs.InfoDialog(label, options, cb);
            },
                (cb) => {
                const homeTabElt = tabs.tabStrip.tabsRoot.querySelector(`li[data-name="home"]`);
                tabs.onActivateTab(homeTabElt);
                callback();
            }
        ]);
    });
}
function updateSystemsAndPlugins() {
    serverSettingsSystems.getRegistry((registry) => {
        const systemsAndPlugins = [];
        for (const systemId in registry.systems) {
            const system = registry.systems[systemId];
            if (!system.isLocalDev && system.localVersion != null && system.version !== system.localVersion)
                systemsAndPlugins.push(systemId);
            for (const authorName in system.plugins) {
                for (const pluginName in system.plugins[authorName]) {
                    const plugin = system.plugins[authorName][pluginName];
                    if (!plugin.isLocalDev && plugin.localVersion != null && plugin.version !== plugin.localVersion)
                        systemsAndPlugins.push(`${systemId}:${authorName}/${pluginName}`);
                }
            }
        }
        if (systemsAndPlugins.length === 0) {
            localServer.start();
            return;
        }
        const label = i18n.t("startup:updateAvailable.systemsAndPlugins", { systemsAndPlugins: systemsAndPlugins.join(", ") });
        const options = {
            validationLabel: i18n.t("common:actions.update"),
            cancelLabel: i18n.t("common:actions.skip")
        };
        new dialogs.ConfirmDialog(label, options, (shouldUpdate) => {
            if (shouldUpdate) {
                openServerSettings_1.default();
                serverSettingsSystems.updateAll(() => { localServer.start(); });
            }
            else {
                localServer.start();
            }
        });
    });
}
