"use strict";
var async = require("async");
var entriesTreeView = require("./sidebar/entriesTreeView");
var sidebar = require("./sidebar");
var tabs = require("./tabs");
var tabsAssets = require("./tabs/assets");
var tabsTools = require("./tabs/tools");
function connect() {
    exports.socket = SupClient.connect(SupClient.query.project, { reconnection: true });
    exports.socket.on("error", onConnectionError);
    exports.socket.on("disconnect", onDisconnected);
    exports.socket.on("welcome", onWelcome);
    exports.socket.on("setProperty:manifest", onSetManifestProperty);
    exports.socket.on("updateIcon:manifest", onUpdateProjectIcon);
    exports.socket.on("add:entries", onEntryAdded);
    exports.socket.on("move:entries", onEntryMoved);
    exports.socket.on("trash:entries", onEntryTrashed);
    exports.socket.on("setProperty:entries", onSetEntryProperty);
    exports.socket.on("save:entries", onEntrySaved);
    exports.socket.on("set:badges", onBadgeSet);
    exports.socket.on("clear:badges", onBadgeCleared);
    exports.socket.on("add:dependencies", onDependenciesAdded);
    exports.socket.on("remove:dependencies", onDependenciesRemoved);
}
exports.connect = connect;
function onConnectionError() {
    var redirect = encodeURIComponent("" + window.location.pathname + window.location.search);
    window.location.replace("/login?redirect=" + redirect);
}
function onDisconnected() {
    SupClient.Dialogs.cancelDialogIfAny();
    exports.entries = null;
    sidebar.disable();
}
function onWelcome(clientId, config) {
    exports.supportsServerBuild = config.supportsServerBuild;
    exports.buildPort = config.buildPort;
    SupClient.fetch("/systems/" + SupCore.system.id + "/plugins.json", "json", function (err, thePluginsInfo) {
        exports.pluginsInfo = thePluginsInfo;
        loadPluginLocales(exports.pluginsInfo.list, function () {
            async.parallel([
                function (cb) { tabsAssets.setup(exports.pluginsInfo.paths.editors, cb); },
                function (cb) { tabsTools.setup(exports.pluginsInfo.paths.tools, cb); }
            ], function (err) {
                if (err)
                    throw err;
                exports.socket.emit("sub", "manifest", null, onManifestReceived);
                exports.socket.emit("sub", "entries", null, onEntriesReceived);
            });
        });
    });
}
function loadPluginLocales(pluginsPaths, cb) {
    var localeFiles = [];
    var pluginsRoot = "/systems/" + SupCore.system.id + "/plugins";
    for (var _i = 0, pluginsPaths_1 = pluginsPaths; _i < pluginsPaths_1.length; _i++) {
        var pluginPath = pluginsPaths_1[_i];
        localeFiles.push({ root: pluginsRoot + "/" + pluginPath, name: "plugin", context: pluginPath });
        localeFiles.push({ root: pluginsRoot + "/" + pluginPath, name: "badges" });
    }
    SupClient.i18n.load(localeFiles, cb);
}
function onManifestReceived(err, manifestPub) {
    exports.manifest = new SupCore.Data.ProjectManifest(manifestPub);
    document.querySelector(".project-name").textContent = manifestPub.name;
    document.title = manifestPub.name + " \u2014 Superpowers";
}
function onEntriesReceived(err, entriesPub, nextEntryId) {
    exports.entries = new SupCore.Data.Entries(entriesPub, nextEntryId);
    sidebar.enable();
    if (SupClient.query.asset != null)
        tabsAssets.open(SupClient.query.asset);
    else if (SupClient.query["tool"] != null)
        tabsTools.open(SupClient.query["tool"]);
}
function onSetManifestProperty(key, value) {
    exports.manifest.client_setProperty(key, value);
    switch (key) {
        case "name":
            document.title = value + " \u2014 Superpowers";
            document.querySelector(".project-name").textContent = value;
            break;
    }
}
function onUpdateProjectIcon() {
    // TODO: Update favicon?
}
function onEntryAdded(entry, parentId, index) {
    exports.entries.client_add(entry, parentId, index);
    var liElt = entriesTreeView.createEntryElement(entry);
    var nodeType = (entry.children != null) ? "group" : "item";
    var parentElt;
    if (parentId != null) {
        parentElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + parentId + "']");
        var parentEntry = exports.entries.byId[parentId];
        var childrenElt = parentElt.querySelector("span.children");
        childrenElt.textContent = "(" + parentEntry.children.length + ")";
    }
    entriesTreeView.widget.insertAt(liElt, nodeType, index, parentElt);
}
function onEntryMoved(id, parentId, index) {
    exports.entries.client_move(id, parentId, index);
    var entryElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + id + "']");
    var oldParentId = entryElt.dataset["parentId"];
    if (oldParentId != null) {
        var oldParentElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + oldParentId + "']");
        var parentEntry = exports.entries.byId[oldParentId];
        var childrenElt = oldParentElt.querySelector("span.children");
        childrenElt.textContent = "(" + parentEntry.children.length + ")";
    }
    var nodeType = (entryElt.classList.contains("group")) ? "group" : "item";
    var parentElt;
    if (parentId != null) {
        parentElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + parentId + "']");
        var parentEntry = exports.entries.byId[parentId];
        var childrenElt = parentElt.querySelector("span.children");
        childrenElt.textContent = "(" + parentEntry.children.length + ")";
    }
    entriesTreeView.widget.insertAt(entryElt, nodeType, index, parentElt);
    if (parentId != null)
        entryElt.dataset["parentId"] = parentId;
    else
        delete entryElt.dataset["parentId"];
    tabsAssets.refreshTabElement(exports.entries.byId[id]);
}
function onEntryTrashed(id) {
    exports.entries.client_remove(id);
    var entryElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + id + "']");
    var oldParentId = entryElt.dataset["parentId"];
    if (oldParentId != null) {
        var oldParentElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + oldParentId + "']");
        var parentEntry = exports.entries.byId[oldParentId];
        var childrenElt = oldParentElt.querySelector("span.children");
        childrenElt.textContent = "(" + parentEntry.children.length + ")";
    }
    entriesTreeView.widget.remove(entryElt);
}
function onSetEntryProperty(id, key, value) {
    exports.entries.client_setProperty(id, key, value);
    var entryElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + id + "']");
    switch (key) {
        case "name":
            entryElt.querySelector(".name").textContent = value;
            var walk_1 = function (entry) {
                tabsAssets.refreshTabElement(entry);
                if (entry.children != null)
                    for (var _i = 0, _a = entry.children; _i < _a.length; _i++) {
                        var child = _a[_i];
                        walk_1(child);
                    }
            };
            walk_1(exports.entries.byId[id]);
            break;
    }
}
function onEntrySaved(id, revisionId, revisionName) {
    exports.entries.client_save(id, revisionId, revisionName);
    var revisionPaneElt = tabs.panesElt.querySelector("[data-asset-id='" + id + "'] .revision-inner-container");
    if (revisionPaneElt == null)
        return;
    var revisions = exports.entries.byId[id].revisions;
    var selectElt = revisionPaneElt.querySelector("select");
    var optionElt = SupClient.html("option", { textContent: revisionName, value: revisionId });
    if (revisions.length === 1) {
        selectElt.appendChild(optionElt);
    }
    else {
        var previousRevisionId = revisions[revisions.length - 2].id;
        var previousRevisionElt = selectElt.querySelector("option[value='" + previousRevisionId + "']");
        selectElt.insertBefore(optionElt, previousRevisionElt);
    }
}
function onBadgeSet(id, newBadge) {
    var badges = exports.entries.badgesByEntryId[id];
    var existingBadge = badges.byId[newBadge.id];
    if (existingBadge != null) {
        existingBadge.type = newBadge.type;
        existingBadge.data = newBadge.data;
    }
    else
        badges.client_add(newBadge, null);
    var badgesElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + id + "'] .badges");
    SupClient.html("span", newBadge.id, { parent: badgesElt, textContent: SupClient.i18n.t("badges:" + newBadge.id) });
}
function onBadgeCleared(id, badgeId) {
    var badges = exports.entries.badgesByEntryId[id];
    badges.client_remove(badgeId);
    var badgeElt = entriesTreeView.widget.treeRoot.querySelector("[data-id='" + id + "'] .badges ." + badgeId);
    badgeElt.parentElement.removeChild(badgeElt);
}
function onDependenciesAdded(id, depIds) {
    for (var _i = 0, depIds_1 = depIds; _i < depIds_1.length; _i++) {
        var depId = depIds_1[_i];
        exports.entries.byId[depId].dependentAssetIds.push(id);
    }
}
function onDependenciesRemoved(id, depIds) {
    for (var _i = 0, depIds_2 = depIds; _i < depIds_2.length; _i++) {
        var depId = depIds_2[_i];
        var dependentAssetIds = exports.entries.byId[depId].dependentAssetIds;
        dependentAssetIds.splice(dependentAssetIds.indexOf(id), 1);
    }
}
