"use strict";
var fs = require("fs");
var path = require("path");
var async = require("async");
var RemoteProjectClient_1 = require("./RemoteProjectClient");
var schemas = require("./schemas");
var migrateProject_1 = require("./migrateProject");
var saveDelay = 60;
var ProjectServer = (function () {
    function ProjectServer(globalIO, projectPath, buildsPath, callback) {
        var _this = this;
        this.projectPath = projectPath;
        this.scheduledSaveCallbacks = {};
        this.nextClientId = 0;
        this.clientsBySocketId = {};
        this.scheduleAssetSave = function (id) {
            var item = _this.data.assets.byId[id];
            if (item == null) {
                SupCore.log("Tried to schedule an asset save for item with id " + id + " but the asset is not loaded.");
                SupCore.log(JSON.stringify(_this.data.entries.byId[id], null, 2));
                SupCore.log((new Error()).stack);
                return;
            }
            var assetPath = path.join(_this.projectPath, "assets/" + _this.data.entries.getStoragePathFromId(id));
            var saveCallback = item.save.bind(item, assetPath);
            _this.scheduleSave(saveDelay, "assets:" + id, saveCallback);
        };
        this.onAssetLoaded = function (assetId, item) {
            item.on("change", function () { _this.scheduleAssetSave(assetId); });
            item.on("edit", function (commandName) {
                var args = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    args[_i - 1] = arguments[_i];
                }
                (_a = _this.io.in("sub:assets:" + assetId)).emit.apply(_a, ["edit:assets", assetId, commandName].concat(args));
                var _a;
            });
            item.on("setBadge", function (badgeId, type, data) { _this.setBadge(assetId, badgeId, type, data); });
            item.on("clearBadge", function (badgeId) { _this.clearBadge(assetId, badgeId); });
            item.on("addDependencies", function (dependencyEntryIds) { _this.addDependencies(assetId, dependencyEntryIds); });
            item.on("removeDependencies", function (dependencyEntryIds) { _this.removeDependencies(assetId, dependencyEntryIds); });
        };
        this.onRoomLoaded = function (roomId, item) {
            var roomPath = path.join(_this.projectPath, "rooms/" + roomId);
            var saveCallback = item.save.bind(item, roomPath);
            item.on("change", function () { _this.scheduleSave(saveDelay, "rooms:" + roomId, saveCallback); });
        };
        this.onResourceLoaded = function (resourceId, item) {
            var resourcePath = path.join(_this.projectPath, "resources/" + resourceId);
            var saveCallback = item.save.bind(item, resourcePath);
            item.on("change", function () { _this.scheduleSave(saveDelay, "resources:" + resourceId, saveCallback); });
            item.on("edit", function (commandName) {
                var args = [];
                for (var _i = 1; _i < arguments.length; _i++) {
                    args[_i - 1] = arguments[_i];
                }
                (_a = _this.io.in("sub:resources:" + resourceId)).emit.apply(_a, ["edit:resources", resourceId, commandName].concat(args));
                var _a;
            });
            item.on("setAssetBadge", function (assetId, badgeId, type, data) { _this.setBadge(assetId, badgeId, type, data); });
            item.on("clearAssetBadge", function (assetId, badgeId) { _this.clearBadge(assetId, badgeId); });
        };
        this.onAddSocket = function (socket) {
            var client = new RemoteProjectClient_1.default(_this, (_this.nextClientId++).toString(), socket);
            _this.clientsBySocketId[socket.id] = client;
        };
        this.scheduleSave = function (minimumSecondsElapsed, callbackName, callback) {
            // this.log(`Scheduling a save: ${callbackName}`);
            var scheduledCallback = _this.scheduledSaveCallbacks[callbackName];
            if (scheduledCallback != null && scheduledCallback.timeoutId != null)
                return;
            var errorCallback = function (err) {
                // this.log(`Save done! ${callbackName}`);
                if (err != null)
                    _this.log("Error in " + callbackName + ":\n" + err);
            };
            if (scheduledCallback == null || scheduledCallback.lastTime <= Date.now() - minimumSecondsElapsed * 1000) {
                _this.scheduledSaveCallbacks[callbackName] = { lastTime: Date.now(), timeoutId: null, callback: null };
                callback(errorCallback);
            }
            else {
                var delay = minimumSecondsElapsed * 1000 - (Date.now() - scheduledCallback.lastTime);
                var timeoutId = setTimeout(function () {
                    callback(errorCallback);
                    scheduledCallback.lastTime = Date.now();
                    scheduledCallback.timeoutId = null;
                    scheduledCallback.callback = null;
                }, delay);
                scheduledCallback.timeoutId = timeoutId;
                scheduledCallback.callback = callback;
            }
        };
        this.onManifestChanged = function () { _this.scheduleSave(saveDelay, "manifest", _this.saveManifest); };
        this.onEntriesChanged = function () { _this.scheduleSave(saveDelay, "entries", _this.saveEntries); };
        this.saveManifest = function (callback) {
            var manifestJSON = JSON.stringify(_this.data.manifest.pub, null, 2);
            fs.writeFile(path.join(_this.projectPath, "manifest.json"), manifestJSON, callback);
        };
        this.saveEntries = function (callback) {
            var entriesJSON = JSON.stringify({ nextEntryId: _this.data.entries.nextId, nodes: _this.data.entries.getForStorage() }, null, 2);
            fs.writeFile(path.join(_this.projectPath, "newEntries.json"), entriesJSON, function () {
                fs.rename(path.join(_this.projectPath, "newEntries.json"), path.join(_this.projectPath, "entries.json"), callback);
            });
        };
        this.data = {
            manifest: null,
            entries: null,
            assets: new SupCore.Data.Assets(this),
            rooms: new SupCore.Data.Rooms(this),
            resources: new SupCore.Data.Resources(this)
        };
        this.data.assets.on("itemLoad", this.onAssetLoaded);
        this.data.rooms.on("itemLoad", this.onRoomLoaded);
        this.data.resources.on("itemLoad", this.onResourceLoaded);
        var loadManifest = function (callback) {
            var done = function (data) {
                try {
                    _this.data.manifest = new SupCore.Data.ProjectManifest(data);
                }
                catch (err) {
                    callback(err);
                    return;
                }
                _this.data.manifest.on("change", _this.onManifestChanged);
                if (_this.data.manifest.migratedFromFormatVersion != null)
                    _this.data.manifest.emit("change");
                _this.system = SupCore.systems[_this.data.manifest.pub.systemId];
                if (_this.system == null) {
                    callback(new Error("The system " + _this.data.manifest.pub.systemId + " is not installed."));
                }
                else {
                    _this.buildsPath = path.join(buildsPath, _this.data.manifest.pub.id);
                    callback(null);
                }
            };
            fs.readFile(path.join(_this.projectPath, "manifest.json"), { encoding: "utf8" }, function (err, manifestJSON) {
                if (err != null) {
                    callback(err);
                    return;
                }
                var manifestData;
                try {
                    manifestData = JSON.parse(manifestJSON);
                }
                catch (err) {
                    callback(err);
                    return;
                }
                try {
                    schemas.validate(manifestData, "projectManifest");
                }
                catch (err) {
                    callback(err);
                    return;
                }
                done(manifestData);
            });
        };
        var setupNextBuildId = function (callback) {
            fs.readdir(_this.buildsPath, function (err, entryIds) {
                if (err != null && err.code !== "ENOENT") {
                    callback(err);
                    return;
                }
                _this.nextBuildId = 0;
                if (entryIds != null) {
                    for (var _i = 0, entryIds_1 = entryIds; _i < entryIds_1.length; _i++) {
                        var entryId = entryIds_1[_i];
                        var entryBuildId = parseInt(entryId, 10);
                        if (isNaN(entryBuildId))
                            continue;
                        _this.nextBuildId = Math.max(entryBuildId + 1, _this.nextBuildId);
                    }
                }
                callback(null);
            });
        };
        var ensureRequiredFoldersExist = function (callback) {
            // These folders might not exist if the project was checked out from source control
            // But we assume they do when saving resources or rooms, so let's recreate them.
            var folders = ["resources", "rooms"];
            async.each(folders, function (folder, cb) {
                fs.mkdir(path.join(_this.projectPath, folder), function (err) {
                    if (err != null && err.code !== "EEXIST") {
                        cb(err);
                        return;
                    }
                    cb();
                });
            }, callback);
        };
        var loadEntries = function (callback) {
            fs.readFile(path.join(_this.projectPath, "entries.json"), { encoding: "utf8" }, function (err, entriesJSON) {
                if (err != null) {
                    callback(err);
                    return;
                }
                var entriesData;
                try {
                    entriesData = JSON.parse(entriesJSON);
                }
                catch (err) {
                    callback(err);
                    return;
                }
                if (_this.data.manifest.migratedFromFormatVersion != null && _this.data.manifest.migratedFromFormatVersion <= 5) {
                    var nextEntryId_1 = 0;
                    var walk_1 = function (node) {
                        var intNodeId = parseInt(node.id, 10);
                        nextEntryId_1 = Math.max(nextEntryId_1, intNodeId);
                        if (node.type == null)
                            for (var _i = 0, _a = node.children; _i < _a.length; _i++) {
                                var childNode = _a[_i];
                                walk_1(childNode);
                            }
                    };
                    for (var _i = 0, entriesData_1 = entriesData; _i < entriesData_1.length; _i++) {
                        var node = entriesData_1[_i];
                        walk_1(node);
                    }
                    nextEntryId_1++;
                    entriesData = { nextEntryId: nextEntryId_1, nodes: entriesData };
                }
                try {
                    schemas.validate(entriesData, "projectEntries");
                }
                catch (err) {
                    callback(err);
                    return;
                }
                _this.data.entries = new SupCore.Data.Entries(entriesData.nodes, entriesData.nextEntryId, _this);
                _this.data.entries.on("change", _this.onEntriesChanged);
                if (_this.data.manifest.migratedFromFormatVersion != null && _this.data.manifest.migratedFromFormatVersion <= 5) {
                    _this.saveEntries(callback);
                }
                else {
                    callback(null);
                }
            });
        };
        // migrate() is called after loadEntries()
        // because some migration code requires the entries to be loaded
        var migrate = function (callback) {
            migrateProject_1.default(_this, callback);
        };
        var serve = function (callback) {
            // Setup the project's namespace
            _this.io = globalIO.of("/project:" + _this.data.manifest.pub.id);
            _this.io.on("connection", _this.onAddSocket);
            callback(null);
        };
        // prepareAssets() and prepareResources() is called after serve()
        // because badges rely on this.io being setup
        var prepareAssets = function (callback) {
            async.each(Object.keys(_this.data.entries.byId), function (assetId, cb) {
                // Ignore folders
                if (_this.data.entries.byId[assetId].type == null) {
                    cb();
                    return;
                }
                _this.data.assets.acquire(assetId, null, function (err, asset) {
                    if (err != null) {
                        cb(err);
                        return;
                    }
                    asset.restore();
                    _this.data.assets.release(assetId, null, { skipUnloadDelay: true });
                    cb();
                });
            }, callback);
        };
        var prepareResources = function (callback) {
            async.each(Object.keys(_this.system.data.resourceClasses), function (resourceId, cb) {
                _this.data.resources.acquire(resourceId, null, function (err, resource) {
                    if (err != null) {
                        cb(err);
                        return;
                    }
                    resource.restore();
                    _this.data.resources.release(resourceId, null, { skipUnloadDelay: true });
                    cb();
                });
            }, callback);
        };
        async.series([
            loadManifest, setupNextBuildId,
            ensureRequiredFoldersExist,
            loadEntries, migrate, serve,
            prepareAssets, prepareResources
        ], callback);
    }
    ProjectServer.prototype.log = function (message) {
        SupCore.log("[" + this.data.manifest.pub.id + " " + this.data.manifest.pub.name + "] " + message);
    };
    ProjectServer.prototype.save = function (callback) {
        var saveCallbacks = [];
        for (var callbackName in this.scheduledSaveCallbacks) {
            var callbackInfo = this.scheduledSaveCallbacks[callbackName];
            if (callbackInfo.timeoutId == null)
                continue;
            clearTimeout(callbackInfo.timeoutId);
            saveCallbacks.push(callbackInfo.callback);
        }
        this.scheduledSaveCallbacks = {};
        async.parallel(saveCallbacks, callback);
    };
    ProjectServer.prototype.moveAssetFolderToTrash = function (trashedAssetFolder, callback) {
        var assetsPath = path.join(this.projectPath, "assets");
        var trashedAssetsPath = path.join(this.projectPath, "trashedAssets");
        fs.mkdir(trashedAssetsPath, function (err) {
            if (err != null && err.code !== "EEXIST")
                throw err;
            var folderPath = path.join(assetsPath, trashedAssetFolder);
            var folderNumber = 0;
            var renameSuccessful = false;
            async.until(function () { return renameSuccessful; }, function (cb) {
                var index = trashedAssetFolder.lastIndexOf("/");
                if (index !== -1)
                    trashedAssetFolder = trashedAssetFolder.slice(index);
                var newFolderPath = path.join(trashedAssetsPath, trashedAssetFolder);
                if (folderNumber > 0)
                    newFolderPath = newFolderPath + " (" + folderNumber + ")";
                fs.rename(folderPath, newFolderPath, function (err) {
                    if (err != null)
                        folderNumber++;
                    else
                        renameSuccessful = true;
                    if (folderNumber > 1000)
                        callback(new Error("Couldn't trash asset: " + trashedAssetFolder));
                    else
                        cb();
                });
            }, callback);
        });
    };
    ProjectServer.prototype.removeRemoteClient = function (socketId) {
        delete this.clientsBySocketId[socketId];
    };
    ProjectServer.prototype.markMissingDependency = function (dependentAssetIds, missingAssetId) {
        for (var _i = 0, dependentAssetIds_1 = dependentAssetIds; _i < dependentAssetIds_1.length; _i++) {
            var dependentAssetId = dependentAssetIds_1[_i];
            var missingAssetIds = [missingAssetId];
            var existingBadge = this.data.entries.badgesByEntryId[dependentAssetId].byId["missingDependencies"];
            if (existingBadge != null) {
                missingAssetIds = missingAssetIds.concat(existingBadge.data.missingAssetIds);
            }
            ;
            this.setBadge(dependentAssetId, "missingDependencies", "error", { missingAssetIds: missingAssetIds });
        }
    };
    ProjectServer.prototype.setBadge = function (assetId, badgeId, type, data) {
        var _this = this;
        // console.log(`setBadge ${assetId} ${badgeId} ${type}`);
        var badges = this.data.entries.badgesByEntryId[assetId];
        var newBadge = { id: badgeId, type: type, data: data };
        var existingBadge = badges.byId[badgeId];
        if (existingBadge != null) {
            existingBadge.type = type;
            existingBadge.data = data;
            this.io.in("sub:entries").emit("set:badges", assetId, newBadge);
            return;
        }
        badges.add(newBadge, null, function (err) {
            _this.io.in("sub:entries").emit("set:badges", assetId, newBadge);
        });
    };
    ProjectServer.prototype.clearBadge = function (assetId, badgeId) {
        var _this = this;
        // console.log(`clearBadge ${assetId} ${badgeId}`);
        var badges = this.data.entries.badgesByEntryId[assetId];
        badges.remove(badgeId, function (err) {
            _this.io.in("sub:entries").emit("clear:badges", assetId, badgeId);
        });
    };
    ProjectServer.prototype.addDependencies = function (assetId, dependencyEntryIds) {
        var addedDependencyEntryIds = [];
        var missingAssetIds = [];
        var assetDependencies = this.data.entries.dependenciesByAssetId[assetId];
        if (assetDependencies == null)
            assetDependencies = this.data.entries.dependenciesByAssetId[assetId] = [];
        for (var _i = 0, dependencyEntryIds_1 = dependencyEntryIds; _i < dependencyEntryIds_1.length; _i++) {
            var depId = dependencyEntryIds_1[_i];
            assetDependencies.push(depId);
            var depEntry = this.data.entries.byId[depId];
            if (depEntry == null) {
                missingAssetIds.push(depId);
                continue;
            }
            var dependentAssetIds = depEntry.dependentAssetIds;
            if (dependentAssetIds.indexOf(assetId) === -1) {
                dependentAssetIds.push(assetId);
                addedDependencyEntryIds.push(depId);
            }
        }
        if (missingAssetIds.length > 0) {
            var existingBadge = this.data.entries.badgesByEntryId[assetId].byId["missingDependencies"];
            if (existingBadge != null)
                missingAssetIds = missingAssetIds.concat(existingBadge.data.missingAssetIds);
            this.setBadge(assetId, "missingDependencies", "error", { missingAssetIds: missingAssetIds });
        }
        if (addedDependencyEntryIds.length > 0) {
            this.io.in("sub:entries").emit("add:dependencies", assetId, addedDependencyEntryIds);
        }
    };
    ProjectServer.prototype.removeDependencies = function (assetId, dependencyEntryIds) {
        var removedDependencyEntryIds = [];
        var missingAssetIds = [];
        var assetDependencies = this.data.entries.dependenciesByAssetId[assetId];
        if (assetDependencies == null)
            assetDependencies = this.data.entries.dependenciesByAssetId[assetId] = [];
        for (var _i = 0, dependencyEntryIds_2 = dependencyEntryIds; _i < dependencyEntryIds_2.length; _i++) {
            var depId = dependencyEntryIds_2[_i];
            assetDependencies.splice(assetDependencies.indexOf(depId), 1);
            var depEntry = this.data.entries.byId[depId];
            if (depEntry == null) {
                missingAssetIds.push(depId);
                continue;
            }
            var dependentAssetIds = depEntry.dependentAssetIds;
            var index = dependentAssetIds.indexOf(assetId);
            if (index !== -1) {
                dependentAssetIds.splice(index, 1);
                removedDependencyEntryIds.push(depId);
            }
        }
        if (missingAssetIds.length > 0) {
            var existingBadge = this.data.entries.badgesByEntryId[assetId].byId["missingDependencies"];
            if (existingBadge != null) {
                for (var _a = 0, missingAssetIds_1 = missingAssetIds; _a < missingAssetIds_1.length; _a++) {
                    var missingAssetId = missingAssetIds_1[_a];
                    var index = existingBadge.data.missingAssetIds.indexOf(missingAssetId);
                    if (index !== -1) {
                        existingBadge.data.missingAssetIds.splice(index, 1);
                    }
                }
                if (existingBadge.data.missingAssetIds.length === 0)
                    this.clearBadge(assetId, "missingDependencies");
                else
                    this.setBadge(assetId, "missingDependencies", "error", existingBadge.data);
            }
        }
        if (removedDependencyEntryIds.length > 0) {
            this.io.in("sub:entries").emit("remove:dependencies", assetId, removedDependencyEntryIds);
        }
        if (assetDependencies.length === 0) {
            delete this.data.entries.dependenciesByAssetId[assetId];
        }
    };
    return ProjectServer;
}());
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = ProjectServer;
