"use strict";
var rimraf = require("rimraf");
var fs = require("fs");
var utils = require("./utils");
function update(systemId, pluginFullName) {
    // Update core
    if (systemId === "core" && pluginFullName == null) {
        var isDevFolder = true;
        try {
            fs.readdirSync(__dirname + "/../../.git");
        }
        catch (err) {
            isDevFolder = false;
        }
        if (isDevFolder)
            utils.emitError("Core is a development version.");
        if (utils.downloadURL != null) {
            updateCore(utils.downloadURL);
            return;
        }
        utils.getRegistry(function (err, registry) {
            if (registry.core.version !== registry.core.localVersion) {
                updateCore(registry.core.downloadURL);
            }
            else {
                console.log("No updates available for the server.");
                process.exit(0);
            }
        });
        return;
    }
    var localSystem = utils.systemsById[systemId];
    if (localSystem == null)
        utils.emitError("System " + systemId + " is not installed.");
    if (pluginFullName == null) {
        // Update system
        if (localSystem.isDev)
            utils.emitError("System " + systemId + " is a development version.");
        if (utils.downloadURL != null) {
            updateSystem(systemId, utils.downloadURL);
            return;
        }
        utils.getRegistry(function (err, registry) {
            if (err)
                utils.emitError("Error while fetching registry:", err.stack);
            var registrySystem = registry.systems[systemId];
            if (registrySystem == null) {
                console.error("System " + systemId + " is not on the registry.");
                utils.listAvailableSystems(registry);
                process.exit(1);
            }
            if (registrySystem.version !== registrySystem.localVersion) {
                updateSystem(systemId, registrySystem.downloadURL);
            }
            else {
                console.log("No updates available for system " + systemId);
                process.exit(0);
            }
        });
    }
    else {
        // Update plugin
        var _a = pluginFullName.split("/"), authorName_1 = _a[0], pluginName_1 = _a[1];
        if (utils.builtInPluginAuthors.indexOf(authorName_1) !== -1)
            utils.emitError("Built-in plugins can not be updated on their own. You must update the system instad.");
        var localPlugin = localSystem.plugins[authorName_1] != null ? localSystem.plugins[authorName_1][pluginName_1] : null;
        if (localPlugin == null)
            utils.emitError("Plugin " + pluginFullName + " is not installed.");
        if (localPlugin.isDev)
            utils.emitError("Plugin " + pluginFullName + " is a development version.");
        if (utils.downloadURL != null) {
            updatePlugin(systemId, pluginFullName, utils.downloadURL);
            return;
        }
        utils.getRegistry(function (err, registry) {
            if (err != null)
                utils.emitError("Error while fetching registry:");
            var registrySystem = registry.systems[systemId];
            if (registrySystem == null) {
                console.error("System " + systemId + " is not on the registry.");
                utils.listAvailableSystems(registry);
                process.exit(1);
            }
            var registryPlugin = registrySystem.plugins[authorName_1] != null ? registrySystem.plugins[authorName_1][pluginName_1] : null;
            if (registryPlugin == null) {
                console.error("Plugin " + pluginFullName + " is not on the registry.");
                utils.listAvailablePlugins(registry, systemId);
                process.exit(1);
            }
            if (registryPlugin.version !== registryPlugin.localVersion) {
                updatePlugin(systemId, pluginFullName, registryPlugin.downloadURL);
            }
            else {
                console.log("No updates available for plugin " + pluginFullName);
                process.exit(0);
            }
        });
    }
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = update;
function updateCore(downloadURL) {
    console.log("Updating the server...");
    var corePath = __dirname + "/../..";
    var newCorePath = __dirname + "/../../../core.new";
    utils.downloadRelease(downloadURL, newCorePath, function (err) {
        if (err != null)
            utils.emitError("Failed to update the core.", err);
        for (var _i = 0, _a = ["client", "node_modules", "public", "server", "SupClient", "SupCore", "LICENSE.txt", "package.json", "README.md"]; _i < _a.length; _i++) {
            var oldItem = _a[_i];
            rimraf.sync(corePath + "/" + oldItem);
        }
        for (var _b = 0, _c = fs.readdirSync(newCorePath); _b < _c.length; _b++) {
            var newItem = _c[_b];
            fs.renameSync(newCorePath + "/" + newItem, corePath + "/" + newItem);
        }
        rimraf.sync(newCorePath);
        console.log("Server successfully updated.");
        process.exit(0);
    });
}
function updateSystem(systemId, downloadURL) {
    console.log("Updating system " + systemId + "...");
    var system = utils.systemsById[systemId];
    var systemPath = utils.systemsPath + "/" + system.folderName;
    var newSystemPath = systemPath + ".new";
    utils.downloadRelease(downloadURL, newSystemPath, function (err) {
        if (err != null)
            utils.emitError("Failed to update the system.", err);
        for (var _i = 0, _a = fs.readdirSync(systemPath); _i < _a.length; _i++) {
            var oldItem = _a[_i];
            if (oldItem === "plugins") {
                for (var _b = 0, _c = fs.readdirSync(systemPath + "/plugins"); _b < _c.length; _b++) {
                    var pluginAuthor = _c[_b];
                    if (utils.builtInPluginAuthors.indexOf(pluginAuthor) === -1)
                        continue;
                    rimraf.sync(systemPath + "/plugins/" + pluginAuthor);
                }
            }
            else {
                rimraf.sync(systemPath + "/" + oldItem);
            }
        }
        for (var _d = 0, _e = fs.readdirSync(newSystemPath); _d < _e.length; _d++) {
            var newItem = _e[_d];
            if (newItem === "plugins") {
                for (var _f = 0, _g = fs.readdirSync(newSystemPath + "/plugins"); _f < _g.length; _f++) {
                    var pluginAuthor = _g[_f];
                    if (utils.builtInPluginAuthors.indexOf(pluginAuthor) === -1)
                        continue;
                    fs.renameSync(newSystemPath + "/plugins/" + pluginAuthor, systemPath + "/plugins/" + pluginAuthor);
                }
            }
            else {
                fs.renameSync(newSystemPath + "/" + newItem, systemPath + "/" + newItem);
            }
        }
        rimraf.sync(newSystemPath);
        console.log("System successfully updated.");
        process.exit(0);
    });
}
function updatePlugin(systemId, pluginFullName, downloadURL) {
    console.log("Updating plugin " + pluginFullName + "...");
    var pluginPath = utils.systemsPath + "/" + utils.systemsById[systemId].folderName + "/plugins/" + pluginFullName;
    var newPluginPath = pluginPath + ".new";
    utils.downloadRelease(downloadURL, newPluginPath, function (err) {
        if (err != null)
            utils.emitError("Failed to update the plugin.", err);
        rimraf.sync(pluginPath);
        fs.renameSync(newPluginPath, pluginPath);
        console.log("Plugin successfully updated.");
        process.exit(0);
    });
}
