"use strict";
var async = require("async");
var path = require("path");
var projectClient;
var subscribersByAssetId = {};
var subscribersByResourceId = {};
var entriesSubscriber;
var entries;
var settings;
var progress = { index: 0, total: 0, errors: 0 };
var statusElt = document.querySelector(".status");
var progressElt = document.querySelector("progress");
var detailsListElt = document.querySelector(".details ol");
function loadPlugins(callback) {
    async.series([
        function (cb) { SupClient.loadScript("/systems/" + SupCore.system.id + "/SupEngine.js", cb); },
        function (cb) { SupClient.loadScript("/systems/" + SupCore.system.id + "/plugins/default/scene/BaseComponentConfig.js", cb); },
        function (cb) {
            async.each(SupCore.system.pluginsInfo.list, function (pluginName, cb) {
                var pluginPath = "/systems/" + SupCore.system.id + "/plugins/" + pluginName;
                async.each(["data", "componentConfigs"], function (name, cb) {
                    SupClient.loadScript(pluginPath + "/bundles/" + name + ".js", cb);
                }, cb);
            }, cb);
        }
    ], function () { callback(); });
}
function build(socket, theSettings) {
    settings = theSettings;
    loadPlugins(function () {
        projectClient = new SupClient.ProjectClient(socket);
        entriesSubscriber = { onEntriesReceived: onEntriesReceived };
        projectClient.subEntries(entriesSubscriber);
    });
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = build;
function onEntriesReceived(theEntries) {
    entries = theEntries;
    projectClient.unsubEntries(entriesSubscriber);
    // Manifest
    projectClient.socket.emit("sub", "manifest", null, onManifestReceived);
    progress.total++;
    // Assets
    entries.walk(function (entry) {
        if (entry.type != null) {
            // Only subscribe to assets that can be exported
            if (SupCore.system.data.assetClasses[entry.type].prototype.clientExport != null) {
                var subscriber = { onAssetReceived: onAssetReceived };
                subscribersByAssetId[entry.id] = subscriber;
                projectClient.subAsset(entry.id, entry.type, subscriber);
                progress.total++;
            }
        }
    });
    // Resources
    for (var resourceId in SupCore.system.data.resourceClasses) {
        // Only subscribe to resources that can be exported
        if (SupCore.system.data.resourceClasses[resourceId].prototype.clientExport != null) {
            var subscriber = { onResourceReceived: onResourceReceived };
            subscribersByResourceId[resourceId] = subscriber;
            projectClient.subResource(resourceId, subscriber);
            progress.total++;
        }
    }
    // TODO: Extra build files
    var systemBuildFiles = ["/SupCore.js"];
    var pluginsInfo = SupCore.system.pluginsInfo;
    var systemPath = "/systems/" + SupCore.system.id;
    for (var _i = 0, _a = pluginsInfo.list; _i < _a.length; _i++) {
        var plugin = _a[_i];
        systemBuildFiles.push(systemPath + "/plugins/" + plugin + "/bundles/components.js");
        systemBuildFiles.push(systemPath + "/plugins/" + plugin + "/bundles/runtime.js");
        systemBuildFiles.push(systemPath + "/plugins/" + plugin + "/bundles/typescriptAPI.js");
    }
    systemBuildFiles.push(systemPath + "/plugins.json");
    systemBuildFiles.push(systemPath + "/index.html");
    systemBuildFiles.push(systemPath + "/index.css");
    systemBuildFiles.push(systemPath + "/images/superpowers-splash.svg");
    systemBuildFiles.push(systemPath + "/SupEngine.js");
    systemBuildFiles.push(systemPath + "/SupRuntime.js");
    var _loop_1 = function(systemBuildFile) {
        var localBuildFile = systemBuildFile;
        if (localBuildFile.indexOf(systemPath) === 0)
            localBuildFile = systemBuildFile.substring(systemPath.length).replace("/", path.sep);
        var outputPath = "" + settings.outputFolder + localBuildFile;
        progress.total++;
        SupClient.fetch(systemBuildFile, "text", function (err, data) {
            if (err != null) {
                progress.errors++;
                SupClient.html("li", { parent: detailsListElt, textContent: SupClient.i18n.t("builds:game.errors.exportFailed", { path: outputPath }) });
                progress.index++;
                updateProgress();
                return;
            }
            var outputDirname = path.dirname(outputPath);
            SupApp.mkdirp(outputDirname, function (err) {
                if (err != null && outputDirname !== settings.outputFolder) {
                    progress.errors++;
                    SupClient.html("li", { parent: detailsListElt, textContent: SupClient.i18n.t("builds:game.errors.exportFailed", { path: outputPath }) });
                    progress.index++;
                    updateProgress();
                    return;
                }
                SupApp.writeFile(outputPath, data, function (err) {
                    if (err != null) {
                        progress.errors++;
                        SupClient.html("li", { parent: detailsListElt, textContent: SupClient.i18n.t("builds:game.errors.exportFailed", { path: outputPath }) });
                    }
                    progress.index++;
                    updateProgress();
                });
            });
        });
    };
    for (var _b = 0, systemBuildFiles_1 = systemBuildFiles; _b < systemBuildFiles_1.length; _b++) {
        var systemBuildFile = systemBuildFiles_1[_b];
        _loop_1(systemBuildFile);
    }
    updateProgress();
}
function onManifestReceived(err, manifestPub) {
    projectClient.socket.emit("unsub", "manifest");
    var exportedProject = { name: manifestPub.name, assets: entries.getForStorage() };
    var json = JSON.stringify(exportedProject, null, 2);
    var projectPath = settings.outputFolder + "/project.json";
    SupApp.writeFile(projectPath, json, { encoding: "utf8" }, function (err) {
        if (err != null) {
            progress.errors++;
            SupClient.html("li", { parent: detailsListElt, textContent: SupClient.i18n.t("builds:game.errors.exportFailed", { path: projectPath }) });
        }
        progress.index++;
        updateProgress();
    });
}
function updateProgress() {
    progressElt.max = progress.total;
    progressElt.value = progress.index;
    if (progress.index < progress.total) {
        statusElt.textContent = SupClient.i18n.t("builds:game.progress", { path: settings.outputFolder, index: progress.index, total: progress.total });
    }
    else if (progress.errors > 0) {
        statusElt.textContent = SupClient.i18n.t("builds:game.doneWithErrors", { path: settings.outputFolder, total: progress.total, errors: progress.errors });
    }
    else {
        statusElt.textContent = SupClient.i18n.t("builds:game.done", { path: settings.outputFolder, total: progress.total });
    }
}
function onAssetReceived(assetId, asset) {
    projectClient.unsubAsset(assetId, subscribersByAssetId[assetId]);
    delete subscribersByAssetId[assetId];
    var outputFolder = settings.outputFolder + "/assets/" + entries.getStoragePathFromId(assetId);
    SupApp.mkdirp(outputFolder, function () {
        asset.clientExport(outputFolder, function (err) {
            if (err != null) {
                progress.errors++;
                SupClient.html("li", { parent: detailsListElt, textContent: SupClient.i18n.t("builds:game.errors.exportFailed", { path: outputFolder }) });
            }
            progress.index++;
            updateProgress();
        });
    });
}
function onResourceReceived(resourceId, resource) {
    projectClient.unsubResource(resourceId, subscribersByResourceId[resourceId]);
    delete subscribersByAssetId[resourceId];
    var outputFolder = settings.outputFolder + "/resources/" + resourceId;
    SupApp.mkdirp(outputFolder, function () {
        resource.clientExport(outputFolder, function (err) {
            if (err != null) {
                progress.errors++;
                SupClient.html("li", { parent: detailsListElt, textContent: SupClient.i18n.t("builds:game.errors.exportFailed", { path: outputFolder }) });
            }
            progress.index++;
            updateProgress();
        });
    });
}
