"use strict";
var ModelRendererUpdater = (function () {
    function ModelRendererUpdater(client, modelRenderer, config, externalSubscriber) {
        var _this = this;
        this.client = client;
        this.modelRenderer = modelRenderer;
        this.externalSubscriber = externalSubscriber;
        this.overrideOpacity = false;
        this.modelAsset = null;
        this.onModelAssetReceived = function (assetId, asset) {
            if (_this.modelRenderer.opacity == null)
                _this.modelRenderer.opacity = asset.pub.opacity;
            _this.prepareMaps(asset.pub.textures, function () {
                _this.modelAsset = asset;
                _this.setModel();
                if (_this.externalSubscriber.onAssetReceived != null)
                    _this.externalSubscriber.onAssetReceived(assetId, asset);
            });
        };
        this.onModelAssetEdited = function (assetId, command) {
            var args = [];
            for (var _i = 2; _i < arguments.length; _i++) {
                args[_i - 2] = arguments[_i];
            }
            var commandFunction = _this.onEditCommands[command];
            if (commandFunction != null)
                commandFunction.apply(_this, args);
            if (_this.externalSubscriber.onAssetEdited != null)
                (_a = _this.externalSubscriber).onAssetEdited.apply(_a, [assetId, command].concat(args));
            var _a;
        };
        this.onEditCommands = {
            setModel: function () {
                _this.setModel();
            },
            setMaps: function (maps) {
                // TODO: Only update the maps that changed, don't recreate the whole model
                _this.prepareMaps(_this.modelAsset.pub.textures, function () {
                    _this.setModel();
                });
            },
            newAnimation: function (animation, index) {
                _this.modelRenderer.updateAnimationsByName();
                _this.playAnimation();
            },
            deleteAnimation: function (id) {
                _this.modelRenderer.updateAnimationsByName();
                _this.playAnimation();
            },
            setAnimationProperty: function (id, key, value) {
                _this.modelRenderer.updateAnimationsByName();
                _this.playAnimation();
            },
            setMapSlot: function (slot, name) { _this.setModel(); },
            deleteMap: function (name) { _this.setModel(); },
            setProperty: function (path, value) {
                switch (path) {
                    case "unitRatio":
                        _this.modelRenderer.setUnitRatio(value);
                        break;
                    case "opacity":
                        if (!_this.overrideOpacity)
                            _this.modelRenderer.setOpacity(value);
                        break;
                }
            }
        };
        this.onModelAssetTrashed = function () {
            _this.modelAsset = null;
            _this.modelRenderer.setModel(null);
        };
        this.onShaderAssetReceived = function (assetId, asset) {
            _this.shaderPub = asset.pub;
            _this.setModel();
        };
        this.onShaderAssetEdited = function (id, command) {
            var args = [];
            for (var _i = 2; _i < arguments.length; _i++) {
                args[_i - 2] = arguments[_i];
            }
            if (command !== "editVertexShader" && command !== "editFragmentShader")
                _this.setModel();
        };
        this.onShaderAssetTrashed = function () {
            _this.shaderPub = null;
            _this.setModel();
        };
        if (this.externalSubscriber == null)
            this.externalSubscriber = {};
        this.modelAssetId = config.modelAssetId;
        this.animationId = config.animationId;
        this.materialType = config.materialType;
        this.shaderAssetId = config.shaderAssetId;
        if (config.castShadow != null)
            this.modelRenderer.castShadow = config.castShadow;
        if (config.receiveShadow != null)
            this.modelRenderer.receiveShadow = config.receiveShadow;
        if (config.overrideOpacity != null) {
            this.overrideOpacity = config.overrideOpacity;
            if (this.overrideOpacity)
                this.modelRenderer.opacity = config.opacity;
        }
        if (config.color != null) {
            var hex = parseInt(config.color, 16);
            this.modelRenderer.color.r = (hex >> 16 & 255) / 255;
            this.modelRenderer.color.g = (hex >> 8 & 255) / 255;
            this.modelRenderer.color.b = (hex & 255) / 255;
        }
        this.modelSubscriber = {
            onAssetReceived: this.onModelAssetReceived,
            onAssetEdited: this.onModelAssetEdited,
            onAssetTrashed: this.onModelAssetTrashed
        };
        this.shaderSubscriber = {
            onAssetReceived: this.onShaderAssetReceived,
            onAssetEdited: this.onShaderAssetEdited,
            onAssetTrashed: this.onShaderAssetTrashed
        };
        if (this.modelAssetId != null)
            this.client.subAsset(this.modelAssetId, "model", this.modelSubscriber);
        if (this.shaderAssetId != null)
            this.client.subAsset(this.shaderAssetId, "shader", this.shaderSubscriber);
    }
    ModelRendererUpdater.prototype.destroy = function () {
        if (this.modelAssetId != null)
            this.client.unsubAsset(this.modelAssetId, this.modelSubscriber);
        if (this.shaderAssetId != null)
            this.client.unsubAsset(this.shaderAssetId, this.shaderSubscriber);
    };
    ModelRendererUpdater.prototype.prepareMaps = function (textures, callback) {
        var textureNames = Object.keys(textures);
        var texturesToLoad = textureNames.length;
        if (texturesToLoad === 0) {
            callback();
            return;
        }
        function onTextureLoaded() {
            texturesToLoad--;
            if (texturesToLoad === 0)
                callback();
        }
        textureNames.forEach(function (key) {
            var image = textures[key].image;
            if (!image.complete)
                image.addEventListener("load", onTextureLoaded);
            else
                onTextureLoaded();
        });
    };
    ModelRendererUpdater.prototype.setModel = function () {
        if (this.modelAsset == null || (this.materialType === "shader" && this.shaderPub == null)) {
            this.modelRenderer.setModel(null);
            return;
        }
        this.modelRenderer.setModel(this.modelAsset.pub, this.materialType, this.shaderPub);
        if (this.animationId != null)
            this.playAnimation();
    };
    ModelRendererUpdater.prototype.playAnimation = function () {
        var animation = this.modelAsset.animations.byId[this.animationId];
        this.modelRenderer.setAnimation((animation != null) ? animation.name : null);
    };
    ModelRendererUpdater.prototype.config_setProperty = function (path, value) {
        switch (path) {
            case "modelAssetId":
                if (this.modelAssetId != null)
                    this.client.unsubAsset(this.modelAssetId, this.modelSubscriber);
                this.modelAssetId = value;
                this.modelAsset = null;
                this.modelRenderer.setModel(null, null);
                if (this.modelAssetId != null)
                    this.client.subAsset(this.modelAssetId, "model", this.modelSubscriber);
                break;
            case "animationId":
                this.animationId = value;
                if (this.modelAsset != null)
                    this.playAnimation();
                break;
            case "castShadow":
                this.modelRenderer.setCastShadow(value);
                break;
            case "receiveShadow":
                this.modelRenderer.threeMesh.receiveShadow = value;
                this.modelRenderer.threeMesh.material.needsUpdate = true;
                break;
            case "overrideOpacity":
                this.overrideOpacity = value;
                this.modelRenderer.setOpacity(value ? null : this.modelAsset.pub.opacity);
                break;
            case "opacity":
                this.modelRenderer.setOpacity(value);
                break;
            case "color":
                var hex = parseInt(value, 16);
                this.modelRenderer.setColor((hex >> 16 & 255) / 255, (hex >> 8 & 255) / 255, (hex & 255) / 255);
                break;
            case "materialType":
                this.materialType = value;
                this.setModel();
                break;
            case "shaderAssetId":
                if (this.shaderAssetId != null)
                    this.client.unsubAsset(this.shaderAssetId, this.shaderSubscriber);
                this.shaderAssetId = value;
                this.shaderPub = null;
                this.modelRenderer.setModel(null);
                if (this.shaderAssetId != null)
                    this.client.subAsset(this.shaderAssetId, "shader", this.shaderSubscriber);
                break;
        }
    };
    return ModelRendererUpdater;
}());
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = ModelRendererUpdater;
